/* $Id: clock_gettime.c 365 2012-12-18 13:26:02Z ken $ */
/*
 * This code was written by KAMADA Ken'ichi and is in the public domain.
 */

#ifdef __MACH__
/*
 * Mac OS X
 * tested on 10.7.3 (Darwin 11.3.0).
 */

#include <mach/clock.h>
#include <mach/mach.h>
#include <errno.h>
#include <time.h>

#include "missing.h"

static int init_clock_ports(void);

static int initialized = 0;
static clock_serv_t clksys, clkcal;

int
clock_gettime(clockid_t clock_id, struct timespec *tp)
{
	mach_timespec_t mts;
	kern_return_t ret;

	if (!initialized) {
		if (init_clock_ports() == -1)
			goto fail;
		initialized = 1;
	}

	switch (clock_id) {
	case CLOCK_REALTIME:
		ret = clock_get_time(clkcal, &mts);
		break;
	case CLOCK_MONOTONIC:
		ret = clock_get_time(clksys, &mts);
		break;
	default:
		goto fail;
	}
	if (ret != KERN_SUCCESS)
		goto fail;
	tp->tv_sec = mts.tv_sec;
	tp->tv_nsec = mts.tv_nsec;
	return 0;
fail:
	errno = EINVAL;
	return -1;
}

static int
init_clock_ports(void)
{
	host_name_port_t host;
	kern_return_t ret;

	host = mach_host_self();
	clksys = MACH_PORT_NULL;
	clkcal = MACH_PORT_NULL;

	ret = host_get_clock_service(host, SYSTEM_CLOCK, &clksys);
	if (ret != KERN_SUCCESS)
		goto fail;
	ret = host_get_clock_service(host, CALENDAR_CLOCK, &clkcal);
	if (ret != KERN_SUCCESS)
		goto fail;
	(void)mach_port_deallocate(mach_task_self(), host);
	return 0;
fail:
	if (clkcal != MACH_PORT_NULL)
		(void)mach_port_deallocate(mach_task_self(), clkcal);
	if (clksys != MACH_PORT_NULL)
		(void)mach_port_deallocate(mach_task_self(), clksys);
	(void)mach_port_deallocate(mach_task_self(), host);
	return -1;
}
#elif 0
/*
 * Generic
 * XXX This does NOT implement the correct semantics of CLOCK_MONOTONIC.
 */

#include <sys/time.h>
#include <errno.h>
#include <time.h>

#include "missing.h"

int
clock_gettime(clockid_t clock_id, struct timespec *tp)
{
	struct timeval tv;

	switch (clock_id) {
	case CLOCK_REALTIME:
	case CLOCK_MONOTONIC:
		if (gettimeofday(&tv, NULL) == -1)
			return -1;
		tp->tv_sec = tv.tv_sec;
		tp->tv_nsec = tv.tv_usec * 1000L;
		return 0;
	default:
		errno = EINVAL;
		return -1;
	}
}
#else
#error "clock_gettime() is not implemented"
#endif
